from pandas import Series, DataFrame
from numpy import array, arange, log10, ndarray
from .expected import _get_expected_digits_
from .constants import digs_dict, rev_digs
from .stats import Z_score
from .checks import _check_num_array_, _check_sign_, _check_decimals_


def _set_N_(len_df, limit_N):
    """"""
    # Assigning to N the superior limit or the lenght of the series
    if limit_N is None or limit_N > len_df:
        return max(1, len_df)
    # Check on limit_N being a positive integer
    else:
        if limit_N < 0 or not isinstance(limit_N, int):
            raise ValueError("limit_N must be None or a positive integer.")
        else:
            return max(1, limit_N)


def get_mantissas(arr):
    """Computes the mantissas, the non-integer part of the log of a number.

    Args:
        arr: array of integers or floats

    Returns:
        Array of floats withe logs mantissas
    """
    log_a = abs(log10(arr))
    return log_a - log_a.astype(int)  # the number - its integer part


def input_data(given):
    """Internalizes and transforms the input data

    Args:
        given: ndarray, Series or tuple with DataFrame and name of the
            column to analyze

    Returns:
        The raw inputed data and the result of its first pre-processing,
            when required.
    """
    if type(given) == Series:
        data = chosen = given
    elif type(given) == ndarray:
        data = given
        chosen = Series(given)
    elif type(given) == tuple:
        if (type(given[0]) != DataFrame) | (type(given[1]) != str):
            raise TypeError('The data tuple must be composed of a pandas '
                            'DataFrame and the name (str) of the chosen '
                            'column, in that order')
        data = given[0]
        chosen = given[0][given[1]]
    else:
        raise TypeError("Wrong data input type. Check docstring.")
    return data, chosen


def set_sign(data, sign="all"):
    """
    """
    sign = _check_sign_(sign)

    if sign == 'all':
        data.seq = data.seq.loc[data.seq != 0]
    elif sign == 'pos':
        data.seq = data.seq.loc[data.seq > 0]
    else:
        data.seq = data.seq.loc[data.seq < 0]

    return data.dropna()


def get_times_10_power(data, decimals=2):
    """"""
    decimals = _check_decimals_(decimals)

    ab = data.seq.abs()

    if data.seq.dtype == 'int':
        data['ZN'] = ab
    else:
        if decimals == 'infer':
            data['ZN'] = ab.astype(str).str\
                .replace('.', '', regex=False)\
                .str.lstrip('0')\
                .str[:5].astype(int)
        else:
            data['ZN'] = (ab * (10 ** decimals)).astype(int)
    return data


def get_digs(data, decimals=2, sign="all"):
    """ 
    """
    df = DataFrame({'seq': _check_num_array_(data)})

    df = set_sign(df, sign=sign)

    df = get_times_10_power(df, decimals=decimals)

    # First digits
    for col in ['F1D', 'F2D', 'F3D']:
        temp = df.ZN.loc[df.ZN >= 10 ** (rev_digs[col] - 1)]
        df[col] = (temp // 10 ** ((log10(temp).astype(int)) -
                                  (rev_digs[col] - 1)))
        # fill NANs with -1, which is a non-usable value for digits,
        # to be discarded later.
        df[col] = df[col].fillna(-1).astype(int)
    # Second digit
    temp_sd = df.loc[df.ZN >= 10]
    df['SD'] = (temp_sd.ZN // 10**((log10(temp_sd.ZN)).astype(int) -
                                   1)) % 10
    df['SD'] = df['SD'].fillna(-1).astype(int)
    # Last two digits
    temp_l2d = df.loc[df.ZN >= 1000]
    df['L2D'] = temp_l2d.ZN % 100
    df['L2D'] = df['L2D'].fillna(-1).astype(int)
    return df


def get_found_proportions(data):
    """
    """
    counts = data.value_counts()
    # get their relative frequencies
    proportions = data.value_counts(normalize=True)
    # crate dataframe from them
    return DataFrame({'Counts': counts, 'Found': proportions}).sort_index()


def join_expect_found_diff(data, digs):
    """
    """
    dd =_get_expected_digits_(digs).join(data).fillna(0)
    # create column with absolute differences
    dd['Dif'] = dd.Found - dd.Expected
    dd['AbsDif'] = dd.Dif.abs()
    return dd


def prepare(data, digs, limit_N=None, simple=False):
    """Transforms the original number sequence into a DataFrame reduced
    by the ocurrences of the chosen digits, creating other computed
    columns
    """
    df = get_found_proportions(data)
    dd = join_expect_found_diff(df, digs)
    if simple:
        del dd['Dif']
        return dd
    else:
        N = _set_N_(len(data), limit_N=limit_N)
        dd['Z_score'] = Z_score(dd, N)
        return N, dd


def subtract_sorted(data):
    """Subtracts the sorted sequence elements from each other, discarding zeros.
    Used in the Second Order test
    """
    temp = data.copy().sort_values(ignore_index=True)
    temp = (temp - temp.shift(1)).dropna()
    return temp.loc[temp != 0]


def prep_to_roll(start, test):
    """Used by the rolling mad and rolling mean, prepares each test and
    respective expected proportions for later application to the Series subset
    """
    if test in [1, 2, 3]:
        start[digs_dict[test]] = start.ZN // 10 ** ((
            log10(start.ZN).astype(int)) - (test - 1))
        start = start.loc[start.ZN >= 10 ** (test - 1)]

        ind = arange(10 ** (test - 1), 10 ** test)
        Exp = log10(1 + (1. / ind))

    elif test == 22:
        start[digs_dict[test]] = (start.ZN // 10 ** ((
            log10(start.ZN)).astype(int) - 1)) % 10
        start = start.loc[start.ZN >= 10]

        Expec = log10(1 + (1. / arange(10, 100)))
        temp = DataFrame({'Expected': Expec, 'Sec_Dig':
                          array(list(range(10)) * 9)})
        Exp = temp.groupby('Sec_Dig').sum().values.reshape(10,)
        ind = arange(0, 10)

    else:
        start[digs_dict[test]] = start.ZN % 100
        start = start.loc[start.ZN >= 1000]

        ind = arange(0, 100)
        Exp = array([1 / 99.] * 100)

    return Exp, ind


def mad_to_roll(arr, Exp, ind):
    """Mean Absolute Deviation used in the rolling function
    """
    prop = arr.value_counts(normalize=True).sort_index()

    if len(prop) < len(Exp):
        prop = prop.reindex(ind).fillna(0)

    return abs(prop - Exp).mean()


def mse_to_roll(arr, Exp, ind):
    """Mean Squared Error used in the rolling function
    """
    temp = arr.value_counts(normalize=True).sort_index()

    if len(temp) < len(Exp):
        temp = temp.reindex(ind).fillna(0)

    return ((temp - Exp) ** 2).mean()



import warnings
from pandas import Series, DataFrame
from numpy import arange, log10, ones, abs, cos, sin, pi, mean
from .constants import confs, digs_dict, sec_order_dict, rev_digs, names, \
    mad_dict, crit_chi2, KS_crit
from .checks import _check_digs_, _check_confidence_, _check_test_, \
    _check_num_array_, _check_high_Z_
from .utils import _set_N_, input_data, prepare, \
    subtract_sorted, prep_to_roll, mad_to_roll, mse_to_roll, \
    get_mantissas
from .expected import _get_expected_digits_ # First, Second, LastTwo
from .viz import _get_plot_args, plot_digs, plot_sum, plot_ordered_mantissas,\
    plot_mantissa_arc_test, plot_roll_mse, plot_roll_mad
from .reports import _inform_, _report_mad_, _report_test_, _deprecate_inform_,\
    _report_mantissa_
from .stats import Z_score, chi_sq, chi_sq_2, kolmogorov_smirnov,\
    kolmogorov_smirnov_2, _bhattacharyya_distance_, _bhattacharyya_coefficient,\
    _kullback_leibler_divergence_


class Base(DataFrame):
    """Internalizes and prepares the data for Analysis.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.`

    Raises:
        TypeError: if not receiving `int` or `float` as input.
    """

    def __init__(self, data, decimals, sign='all', sec_order=False):

        DataFrame.__init__(self, {'seq': data})

        if (self.seq.dtype != 'float') & (self.seq.dtype != 'int'):
            raise TypeError("The sequence dtype was neither int nor "
                            "float. Convert it to whether int of float, "
                            "and try again.")

        if sign == 'all':
            self.seq = self.seq.loc[self.seq != 0]
        elif sign == 'pos':
            self.seq = self.seq.loc[self.seq > 0]
        else:
            self.seq = self.seq.loc[self.seq < 0]

        self.dropna(inplace=True)

        ab = self.seq.abs()

        if self.seq.dtype == 'int':
            self['ZN'] = ab
        else:
            if decimals == 'infer':
                self['ZN'] = ab.astype(str).str\
                               .replace('.', '', regex=False)\
                               .str.lstrip('0')\
                               .str[:5].astype(int)
            else:
                self['ZN'] = (ab * (10 ** decimals)).astype(int)
        # First digits
        for col in ['F1D', 'F2D', 'F3D']:
            temp = self.ZN.loc[self.ZN >= 10 ** (rev_digs[col] - 1)]
            self[col] = (temp // 10 ** ((log10(temp).astype(int)) -
                                        (rev_digs[col] - 1)))
            # fill NANs with -1, which is a non-usable value for digits,
            # to be discarded later.
            self[col] = self[col].fillna(-1).astype(int)
        # Second digit
        temp_sd = self.loc[self.ZN >= 10]
        self['SD'] = (temp_sd.ZN // 10**((log10(temp_sd.ZN)).astype(int) -
                                         1)) % 10
        self['SD'] = self['SD'].fillna(-1).astype(int)
        # Last two digits
        temp_l2d = self.loc[self.ZN >= 1000]
        self['L2D'] = temp_l2d.ZN % 100
        self['L2D'] = self['L2D'].fillna(-1).astype(int)


class Test(DataFrame):
    """Transforms the original number sequence into a DataFrame reduced
    by the ocurrences of the chosen digits, creating other computed
    columns

    Args:
        base: The Base object with the data prepared for Analysis
        digs: Tells which test to perform: 1: first digit; 2: first two digits;
            3: furst three digits; 22: second digit; -2: last two digits.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show.
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.

    Attributes:
        N: Number of records in the sample to consider in computations
        ddf: Degrees of Freedom to look up for the critical chi-square value
        chi_square: Chi-square statistic for the given test
        KS: Kolmogorov-Smirnov statistic for the given test
        MAD: Mean Absolute Deviation for the given test
        confidence: Confidence level to consider when setting some critical values
        digs (int): numerical representation of the test at hand. 1: F1D; 2: F2D;
            3: F3D; 22: SD; -2: L2D.
        sec_order (bool): True if the test is a Second Order one
    """

    def __init__(self, base, digs, confidence, limit_N=None, sec_order=False):
        # create a separated Expected distributions object
        super(Test, self).__init__(_get_expected_digits_(digs))
        # create column with occurrences of the digits in the base
        self['Counts'] = base[digs_dict[digs]].value_counts()
        # create column with relative frequencies
        self['Found'] = base[digs_dict[digs]].value_counts(normalize=True)
        self.fillna(0, inplace=True)
        # create column with absolute differences
        self['Dif'] = self.Found - self.Expected
        self['AbsDif'] = self.Dif.abs()
        self.N = _set_N_(len(base), limit_N)
        self['Z_score'] = Z_score(self, self.N)
        self.ddf = len(self) - 1
        self.chi_square = chi_sq_2(self)
        self.KS = kolmogorov_smirnov_2(self)
        self.MAD = self.AbsDif.mean()
        self.MSE = (self.AbsDif ** 2).mean()
        self.bhattacharyya_coefficient = _bhattacharyya_coefficient(
            self.Found.values, self.Expected.values)
        self.bhattacharyya_distance = _bhattacharyya_distance_(
            self.Found.values, self.Expected.values)
        self.kullback_leibler_divergence = _kullback_leibler_divergence_(
            self.Found.values, self.Expected.values)
        self.confidence = confidence
        self.digs = digs
        self.sec_order = sec_order

        if sec_order:
            self.name = names[sec_order_dict[digs]]
        else:
            self.name = names[digs_dict[digs]]

    def update_confidence(self, new_conf, check=True):
        """Sets a new confidence level for the Benford object, so as to be used to
        produce critical values for the tests

        Args:
            new_conf: new confidence level to draw lower and upper limits when
                plotting and to limit the top deviations to show, as well as to
                calculate critical values for the tests' statistics.
            check: checks the value provided for the confidence. Defaults to True
        """
        if check:
            self.confidence = _check_confidence_(new_conf)
        else:
            self.confidence = new_conf

    @property
    def critical_values(self):
        """dict: a dictionary with the critical values for the test at hand,
            according to the current confidence level."""
        return {'Z': confs[self.confidence],
                'KS': KS_crit[self.confidence] / (self.N ** 0.5),
                'chi2': crit_chi2[self.ddf][self.confidence],
                'MAD': mad_dict[self.digs]
                }

    def show_plot(self, save_plot=None, save_plot_kwargs=None):
        """Draws the test plot.
        
        Args:
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        x, figsize, text_x = _get_plot_args(self.digs)
        plot_digs(self, x=x, y_Exp=self.Expected, y_Found=self.Found,
                    N=self.N, figsize=figsize, conf_Z=confs[self.confidence],
                    text_x=text_x, save_plot=save_plot, save_plot_kwargs=save_plot_kwargs
                    )

    def report(self, high_Z='pos', show_plot=True,
               save_plot=None, save_plot_kwargs=None):
        """Handles the report especific to the test, considering its statistics
        and according to the current confidence level.

        Args:
            high_Z (int): chooses which Z scores to be used when displaying results,
                according to the confidence level chosen. Defaluts to 'pos',
                which will highlight only values higher than the expexted
                frequencies; 'all' will highlight both extremes (positive and
                negative); and an integer, which will use the first n entries,
                positive and negative, regardless of whether Z is higher than
                the critical value or not.
            show_plot: calls the show_plot method, to draw the test plot
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        high_Z = _check_high_Z_(high_Z)
        _report_test_(self, high_Z, self.critical_values)
        if show_plot:
            self.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)


class Summ(DataFrame):
    """Gets the base object and outputs a Summation test object

    Args:
       base: The Base object with the data prepared for Analysis
       test: The test for which to compute the summation
    """

    def __init__(self, base, test):
        super(Summ, self).__init__(base.abs()
                                   .groupby(test)[['seq']]
                                   .sum())
        self['Percent'] = self.seq / self.seq.sum()
        self.columns.values[0] = 'Sum'
        self.expected = 1 / len(self)
        self['AbsDif'] = (self.Percent - self.expected).abs()
        self.index = self.index.astype(int)
        #: Mean Absolute Deviation for the test
        self.MAD = self.AbsDif.mean()
        self.MSE = (self.AbsDif ** 2).mean()
        #: Confidence level to consider when setting some critical values
        self.confidence = None
        # (int): numerical representation of the test at hand
        self.digs = rev_digs[test]
        # (str): the name of the Summation test.
        self.name = names[f'{test}_Summ']

    def show_plot(self, save_plot=None, save_plot_kwargs=None):
        """Draws the Summation test plot
        
        Args:
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        figsize=(2 * (self.digs ** 2 + 5), 1.5 * (self.digs ** 2 + 5))
        plot_sum(self, figsize, self.expected,
                 save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    
    def report(self, high_diff=None, show_plot=True,
               save_plot=None, save_plot_kwargs=None):
        """Gives the report on the Summation test.

        Args:
            high_diff: Number of records to show after ordering by the absolute
                differences between the found and the expected proportions
            show_plot: calls the show_plot method, to draw the Summation test plot
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        _report_test_(self, high_diff)
        if show_plot:
            self.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)


class Mantissas(object):
    """Computes and holds the mantissas of the logarithms of the records

    Args:
        data: sequence to compute mantissas from. numpy 1D array, pandas
            Series of pandas DataFrame column.
    """

    def __init__(self, data):

        data = Series(_check_num_array_(data))
        data = data.dropna().loc[data != 0].abs()
        #: (DataFrame): pandas DataFrame with the mantissas
        self.data = DataFrame({'Mantissa': get_mantissas(data.abs())})
        # (dict): Dictionary with the mantissas statistics
        self.stats = {'Mean': self.data.Mantissa.mean(),
                      'Var': self.data.Mantissa.var(),
                      'Skew': self.data.Mantissa.skew(),
                      'Kurt': self.data.Mantissa.kurt()}

    def report(self, show_plot=True, save_plot=None, save_plot_kwargs=None):
        """Displays the Mantissas test stats.

        Args:
            show_plot: shows the Ordered Mantissas plot and the Arc Test plot.
                Defaults to True.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        _report_mantissa_(self.stats)

        if show_plot:
            self.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
            self.arc_test(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)

    def show_plot(self, figsize=(12, 6), save_plot=None, save_plot_kwargs=None):
        """Plots the ordered mantissas and a line with the expected
        inclination.

        Args:
            figsize (tuple): figure size dimensions
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        plot_ordered_mantissas(self.data.Mantissa, figsize=figsize,
                               save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)

    def arc_test(self, decimals=2, grid=True, figsize=12,
                 save_plot=None, save_plot_kwargs=None):
        """Adds two columns to Mantissas's DataFrame equal to their "X" and "Y"
        coordinates, plots its to a scatter plot and calculates the gravity
        center of the circle.

        Args:
            decimals: number of decimal places for displaying the gravity center.
                Defaults to 2.
            grid: show grid of the plot. Defaluts to True.
            figsize (int): size of the figure to be displayed. Since it is a square,
                there is no need to provide a tuple, like is usually the case with
                matplotlib.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        if self.stats.get('gravity_center') is None:
            self.data['mant_x'] = cos(2 * pi * self.data.Mantissa)
            self.data['mant_y'] = sin(2 * pi * self.data.Mantissa)
            self.stats['gravity_center'] = (self.data.mant_x.mean(),
                                            self.data.mant_y.mean())
        
        plot_mantissa_arc_test(self.data, self.stats, decimals=decimals, 
                               grid=grid, figsize=figsize,
                               save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)


class Benford(object):
    """Initializes a Benford Analysis object and computes the proportions for
    the digits. The tets dataFrames are atributes, i.e., obj.F1D is the First
    Digit DataFrame, the obj.F2D,the First Two Digits one, and so one, F3D for
    First Three Digits, SD for Second  Digit and L2D for Last Two Digits.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a tuple with a pandas DataFrame and the name (str)
            of the chosen column. Values must be integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show, as well as to
            calculate critical values for the tests' statistics. Defaults to 95.
        sec_order: runs the Second Order tests, which are the Benford's tests
            performed on the differences between the ordered sample (a value minus
            the one before it, and so on). If the original series is Benford-
            compliant, this new sequence should aldo follow Beford. The Second
            Order can also be called separately, through the method sec_order().
        summation: creates the Summation DataFrames for the First, First Two, and
            First Three Digits. The summation tests can also be called separately,
            through the method summation().
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.
        verbose: gives some information about the data and the registries used
            and discarded for each test.

    Attributes:
        data: the raw data provided for the analysis
        chosen: the column of the DataFrame to be analysed or the data itself
        sign (str): which number sign(s) to include in the analysis
        confidence: current confidence level
        limit_N (int): sample size to use in computations
        verbose (bool): verbose or not
        base: the Base, pre-processed object
        tests (:obj:`list` of :obj:`str`): keeps track of the tests the
            instance has
    """

    def __init__(self, data, decimals=2, sign='all', confidence=95,
                 mantissas=False, sec_order=False, summation=False,
                 limit_N=None, verbose=True):
        self.data, self.chosen = input_data(data)
        self.decimals = decimals
        self.sign = sign
        self.confidence = _check_confidence_(confidence)
        self.limit_N = limit_N
        self.verbose = verbose
        self.base = Base(self.chosen, decimals, sign)
        self.tests = []

        # Create a DatFrame for each Test
        for key, val in digs_dict.items():
            test = Test(self.base.loc[self.base[val] != -1],
                        digs=key, confidence=self.confidence,
                        limit_N=self.limit_N)
            setattr(self, val, test)
            self.tests.append(val)
        # dict with the numbers of discarded entries for each test column
        self._discarded = {key: val for (key, val) in
                           zip(digs_dict.values(),
                               [len(self.base[col].loc[self.base[col] == -1])
                                for col in digs_dict.values()])}

        if self.verbose:
            print('\n', ' Benford Object Instantiated '.center(50, '#'), '\n')
            print(f'Initial sample size: {len(self.chosen)}.\n')
            print(f'Test performed on {len(self.base)} registries.\n')
            print(
                f'Number of discarded entries for each test:\n{self._discarded}')

        if mantissas:
            self.mantissas()

        if sec_order:
            self.sec_order()

        if summation:
            self.summation()

    def update_confidence(self, new_conf, tests=None):
        """Sets (a) new confidence level(s) for the Benford object, so as to be
        used to produce critical values for the tests.

        Args:
            new_conf: new confidence level to draw lower and upper limits when
                plotting and to limit the top deviations to show, as well as to
                calculate critical values for the tests' statistics.
            tests (:obj:`list` of :obj:`str`): list of tests names (strings) to
                have their confidence updated. If only one, provide a one-element
                list, like ['F1D']. Defauts to None, in which case it will use
                the instance .test list attribute.

        Raises:
            ValueError: if the test argument is not a `list` or `None`.
        """
        self.confidence = _check_confidence_(new_conf)
        if tests is None:
            tests = self.tests
        else:
            if not isinstance(tests, list):
                raise ValueError('tests must be a list or None.')
        for test in tests:
            try:
                getattr(self, test).update_confidence(
                    self.confidence, check=False)
            except AttributeError:
                if test in ['Mantissas', 'F1D_Summ', 'F2D_Summ', 'F3D_Summ']:
                    pass
                else:
                    print(
                        f"{test} not in Benford instance tests - review test's name.")
                    pass

    @property
    def all_confidences(self):
        """dict: a dictionary with a confidence level for each computed tests,
        when applicable."""
        con_dic = {}
        for key in self.tests:
            try:
                con_dic[key] = getattr(self, key).confidence
            except AttributeError:
                pass
        return con_dic

    def mantissas(self):
        """Adds a Mantissas object to the tests, with all its statistics and
        plotting capabilities.
        """
        self.Mantissas = Mantissas(self.base.seq)
        self.tests.append('Mantissas')
        if self.verbose:
            print('\nAdded Mantissas test.')

    def sec_order(self):
        """Runs the Second Order tests, which are the Benford's tests
        performed on the differences between the ordered sample (a value minus
        the one before it, and so on). If the original series is Benford-
        compliant, this new sequence should aldo follow Beford. The Second
        Order can also be called separately, through the method sec_order().
        """
        #: Base instance of the differences between the ordered sample
        self.base_sec = Base(subtract_sorted(self.chosen),
                             decimals=self.decimals, sign=self.sign)
        for key, val in digs_dict.items():
            test = Test(self.base_sec.loc[self.base_sec[val] != -1],
                        digs=key, confidence=self.confidence,
                        limit_N=self.limit_N, sec_order=True)
            setattr(self, sec_order_dict[key], test)
            self.tests.append(f'{val}_sec')
            # No need to populate crit_vals dict, since they are the
            # same and do not depend on N
            self._discarded_sec = {key: val for (key, val) in zip(
                                   sec_order_dict.values(),
                                   [sum(self.base_sec[col] == -1) for col in
                                    digs_dict.values()])}
        if self.verbose:
            print(f'\nSecond order tests run in {len(self.base_sec)} '
                  'registries.\n\nNumber of discarded entries for second order'
                  f' tests:\n{self._discarded_sec}')

    def summation(self):
        """Creates Summation test DataFrames from Base object"""
        for test in ['F1D', 'F2D', 'F3D']:
            t = f'{test}_Summ'
            setattr(self, t, Summ(self.base, test))
            self.tests.append(t)

        if self.verbose:
            print('\nAdded Summation DataFrames to F1D, F2D and F3D Tests.')


class Source(DataFrame):
    """Prepares the data for Analysis. pandas DataFrame subclass.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.
        sec_order: choice for the Second Order Test, which cumputes the
            differences between the ordered entries before running the Tests.
        verbose (bool): tells the number of registries that are being subjected to
            the analysis; defaults to True.

    Raises:
        ValueError: if the `sign` arg is not in ['all', 'pos', 'neg']
        TypeError: if not receiving `int` or `float` as input.
    """

    def __init__(self, data, decimals=2, sign='all', sec_order=False,
                 verbose=True, inform=None):

        if sign not in ['all', 'pos', 'neg']:
            raise ValueError("The -sign- argument must be "
                             "'all','pos' or 'neg'.")

        DataFrame.__init__(self, {'seq': data})

        if self.seq.dtype != 'float' and self.seq.dtype != 'int':
            raise TypeError('The sequence dtype was neither int nor float.\n'
                            'Convert it to whether int or float, and try again.')

        if sign == 'pos':
            self.seq = self.seq.loc[self.seq > 0]
        elif sign == 'neg':
            self.seq = self.seq.loc[self.seq < 0]
        else:
            self.seq = self.seq.loc[self.seq != 0]

        self.dropna(inplace=True)
        #: (bool): verbose or not
        self.verbose = _deprecate_inform_(verbose, inform)
        if self.verbose:
            print(f"\nInitialized sequence with {len(self)} registries.")

        if sec_order:
            self.seq = subtract_sorted(self.seq.copy())
            self.dropna(inplace=True)
            self.reset_index(inplace=True)
            if verbose:
                print('Second Order Test. Initial series reduced '
                      f'to {len(self.seq)} entries.')

        ab = self.seq.abs()

        if self.seq.dtype == 'int':
            self['ZN'] = ab
        else:
            if decimals == 'infer':
                # There is some numerical issue with Windows that required
                # implementing it differently (and slower)
                self['ZN'] = ab.astype(str)\
                               .str.replace('.', '', regex=False)\
                               .str.lstrip('0').str[:5]\
                               .astype(int)
            else:
                self['ZN'] = (ab * (10 ** decimals)).astype(int)

    def mantissas(self, report=True, show_plot=True, figsize=(15, 8),
                  save_plot=None, save_plot_kwargs=None):
        """Calculates the mantissas, their mean and variance, and compares them
        with the mean and variance of a Benford's sequence.

        Args:
            report: prints the mamtissas mean, variance, skewness and kurtosis
                for the sequence studied, along with reference values.
            show_plot: plots the ordered mantissas and a line with the expected
                inclination. Defaults to True.
            figsize: tuple that sets the figure dimensions.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        self['Mant'] = get_mantissas(self.seq.abs())
        if report:
            p = self[['seq', 'Mant']]
            p = p.loc[p.seq > 0].sort_values('Mant')
            print(f"The Mantissas MEAN is {p.Mant.mean()}. Ref: 0.5.")
            print(f"The Mantissas VARIANCE is {p.Mant.var()}. Ref: 0.083333.")
            print(f"The Mantissas SKEWNESS is {p.Mant.skew()}. \tRef: 0.")
            print(f"The Mantissas KURTOSIS is {p.Mant.kurt()}. \tRef: -1.2.")

        if show_plot:
            plot_ordered_mantissas(self.Mant, figsize=figsize,
                                   save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)

    def first_digits(self, digs, confidence=None, high_Z='pos',
                     limit_N=None, MAD=False, MSE=False, chi_square=False,
                     KS=False, show_plot=True, save_plot=None, save_plot_kwargs=None,
                     simple=False, bhat_coeff = False, bhat_dist=False,
                     kl_diverg=False, ret_df=False):
        """Performs the Benford First Digits test with the series of
        numbers provided, and populates the mapping dict for future
        selection of the original series.

        Args:
            digs (int): number of first digits to consider. Must be 1 (first digit),
                2 (first two digits) or 3 (first three digits).
            verbose (bool): tells the number of registries that are being subjected to
                the analysis; defaults to True
            confidence (int, float): confidence level to draw lower and upper limits when
                plotting and to limit the top deviations to show, as well as to
                calculate critical values for the tests' statistics. Defaults to None.
            high_Z (int): chooses which Z scores to be used when displaying results,
                according to the confidence level chosen. Defaluts to 'pos',
                which will highlight only values higher than the expexted
                frequencies; 'all' will highlight both extremes (positive and
                negative); and an integer, which will use the first n entries,
                positive and negative, regardless of whether Z is higher than
                the confidence or not.
            limit_N (int): sets a limit to N as the sample size for the calculation of
                the Z scores if the sample is too big. Defaults to None.
            MAD (bool): calculates the Mean Absolute Difference between the
                found and the expected distributions; defaults to False.
            MSE (bool): calculates the Mean Square Error of the sample; defaults to
                False.
            bhat_coeff (bool): computes the Bhattacharyya Coefficient between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            bhat_dist (bool): calculates the Bhattacharyya Distance between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            kl_diverg (bool): calculates the Kulback-Laibler Divergence between
                the found and the expected (Benford) digits distribution;
                defaults to False
            show_plot (bool): draws the test plot. Defaults to True.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
            ret_df: returns the test DataFrame. Defaults to False. True if run by
                the test function.

        Returns:
            DataFrame with the Expected and Found proportions, and the Z scores of
                the differences
        """
        # Check on the possible values for confidence levels
        confidence = _check_confidence_(confidence)
        # Check on possible digits
        _check_test_(digs)

        temp = self.loc[self.ZN >= 10 ** (digs - 1)]
        temp[digs_dict[digs]] = (temp.ZN // 10 ** ((log10(temp.ZN).astype(
                                                   int)) - (digs - 1))).astype(
                                                       int)
        n, m = 10 ** (digs - 1), 10 ** (digs)
        x = arange(n, m)

        if simple:
            self.verbose = False
            show_plot = False
            df = prepare(temp[digs_dict[digs]], digs, limit_N=limit_N,
                         simple=True)
        else:
            N, df = prepare(temp[digs_dict[digs]], digs, limit_N=limit_N,
                            simple=False)

        if self.verbose:
            print(f"\nTest performed on {len(temp)} registries.\n"
                  f"Discarded {len(self) - len(temp)} records < {10 ** (digs - 1)}"
                  " after preparation.")
            if confidence is not None:
                _inform_(df, high_Z=high_Z, conf=confs[confidence])

        # Mean absolute difference
        if MAD:
            self.MAD = df.AbsDif.mean()
            if self.verbose:
                _report_mad_(digs, self.MAD)

        # Mean Square Error
        if MSE:
            self.MSE = (df.AbsDif ** 2).mean()

        # Chi-square statistic
        if chi_square:
            self.chi_square = chi_sq(df, ddf=len(df) - 1,
                                     confidence=confidence,
                                     verbose=self.verbose)
        # KS test
        if KS:
            self.KS = kolmogorov_smirnov(df, confidence=confidence, N=len(temp),
                                         verbose=self.verbose)

        if bhat_coeff:
            self.bhat_coeff = _bhattacharyya_coefficient(
                                df.Found.values, df.Expected.values)

        if bhat_dist:
            self.bhat_dist = _bhattacharyya_distance_(
                                df.Found.values, df.Expected.values)
        
        if kl_diverg:
            self.kl_diverg = _kullback_leibler_divergence_(
                                df.Found.values, df.Expected.values)

        # Plotting the expected frequncies (line) against the found ones(bars)
        if show_plot:
            plot_digs(df, x=x, y_Exp=df.Expected, y_Found=df.Found, N=N,
                       figsize=(2 * (digs ** 2 + 5), 1.5 * (digs ** 2 + 5)),
                       conf_Z=confs[confidence], save_plot=save_plot,
                       save_plot_kwargs=save_plot_kwargs)
        if ret_df:
            return df

    def second_digit(self, confidence=None, high_Z='pos',
                     limit_N=None, MAD=False, MSE=False, chi_square=False,
                     KS=False, bhat_coeff=False, bhat_dist=False, kl_diverg=False,
                     show_plot=True, save_plot=None, save_plot_kwargs=None,
                     simple=False, ret_df=False):
        """Performs the Benford Second Digit test with the series of
        numbers provided.

        Args:
            verbose (bool): tells the number of registries that are being subjected to
                the analysis; defaults to True
            MAD (bool): calculates the Mean Absolute Difference between the
                found and the expected distributions; defaults to False.
            confidence (int, float): confidence level to draw lower and upper limits when
                plotting and to limit the top deviations to show, as well as to
                calculate critical values for the tests' statistics. Defaults to None.
            high_Z (int): chooses which Z scores to be used when displaying results,
                according to the confidence level chosen. Defaluts to 'pos',
                which will highlight only values higher than the expexted
                frequencies; 'all' will highlight both extremes (positive and
                negative); and an integer, which will use the first n entries,
                positive and negative, regardless of whether Z is higher than
                the confidence or not.
            limit_N (int): sets a limit to N as the sample size for the calculation of
                the Z scores if the sample is too big. Defaults to None.
            MSE (bool): calculates the Mean Square Error of the sample; defaults to
                False.
            bhat_coeff (bool): computes the Bhattacharyya Coefficient between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            bhat_dist (bool): calculates the Bhattacharyya Distance between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            kl_diverg (bool): calculates the Kulback-Laibler Divergence between
                the found and the expected (Benford) digits distribution;
                defaults to False
            show_plot (bool): draws the test plot.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
            ret_df: returns the test DataFrame. Defaults to False. True if run by
                the test function.

        Returns:
            DataFrame with the Expected and Found proportions, and the Z scores of
                the differences
        """
        confidence = _check_confidence_(confidence)

        conf = confs[confidence]

        temp = self.loc[self.ZN >= 10, :]
        temp['SD'] = (temp.ZN // 10 ** ((log10(temp.ZN)).astype(
                      int) - 1)) % 10

        if simple:
            self.verbose = False
            show_plot = False
            df = prepare(temp['SD'], 22, limit_N=limit_N, simple=True)
        else:
            N, df = prepare(temp['SD'], 22, limit_N=limit_N, simple=False)

        if self.verbose:
            print(f"\nTest performed on {len(temp)} registries.\nDiscarded "
                  f"{len(self) - len(temp)} records < 10 after preparation.")
            if confidence is not None:
                _inform_(df, high_Z, conf)

        # Mean absolute difference
        if MAD:
            self.MAD = df.AbsDif.mean()
            if self.verbose:
                _report_mad_(22, self.MAD)
        # Mean Square Error
        if MSE:
            self.MSE = (df.AbsDif ** 2).mean()

        # Chi-square statistic
        if chi_square:
            self.chi_square = chi_sq(df, ddf=9, confidence=confidence,
                                     verbose=self.verbose)
        # KS test
        if KS:
            self.KS = kolmogorov_smirnov(df, confidence=confidence, N=len(temp),

                                         verbose=self.verbose)
        if bhat_coeff:
            self.bhat_coeff = _bhattacharyya_coefficient(
                                df.Found.values, df.Expected.values)

        if bhat_dist:
            self.bhat_dist = _bhattacharyya_distance_(
                                df.Found.values, df.Expected.values
                            )
        
        if kl_diverg:
            self.kl_diverg = _kullback_leibler_divergence_(
                                df.Found.values, df.Expected.values
                            )

        # Plotting the expected frequncies (line) against the found ones(bars)
        if show_plot:
            plot_digs(df, x=arange(0, 10), y_Exp=df.Expected,
                       y_Found=df.Found, N=N, figsize=(10, 6), conf_Z=conf,
                       save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
        if ret_df:
            return df

    def last_two_digits(self, confidence=None, high_Z='pos',
                        limit_N=None, MAD=False, MSE=False, chi_square=False,
                        KS=False, bhat_coeff=False, bhat_dist=False, kl_diverg=False,
                        show_plot=True, save_plot=None, save_plot_kwargs=None,
                        simple=False, ret_df=False):
        """Performs the Benford Last Two Digits test with the series of
        numbers provided.

        Args:
            verbose (bool): tells the number of registries that are being subjected to
                the analysis; defaults to True
            MAD (bool): calculates the Mean Absolute Difference between the
                found and the expected distributions; defaults to False.
            confidence (int, float): confidence level to draw lower and upper limits when
                plotting and to limit the top deviations to show, as well as to
                calculate critical values for the tests' statistics. Defaults to None.
            high_Z (int): chooses which Z scores to be used when displaying results,
                according to the confidence level chosen. Defaluts to 'pos',
                which will highlight only values higher than the expexted
                frequencies; 'all' will highlight both extremes (positive and
                negative); and an integer, which will use the first n entries,
                positive and negative, regardless of whether Z is higher than
                the confidence or not.
            limit_N (int): sets a limit to N as the sample size for the calculation of
                the Z scores if the sample is too big. Defaults to None.
            MSE (bool): calculates the Mean Square Error of the sample; defaults to
                False.
            bhat_coeff (bool): computes the Bhattacharyya Coefficient between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            bhat_dist (bool): calculates the Bhattacharyya Distance between
                the found and the expected (Benford) digits distribution; defaults
                to Fasle
            kl_diverg (bool): calculates the Kulback-Laibler Divergence between
                the found and the expected (Benford) digits distribution;
                defaults to False
            show_plot (bool): draws the test plot.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        
        Returns:
            DataFrame with the Expected and Found proportions, and the Z scores of
                the differences
        """
        confidence = _check_confidence_(confidence)
        conf = confs[confidence]

        temp = self.loc[self.ZN >= 1000]
        temp['L2D'] = temp.ZN % 100

        if simple:
            self.verbose = False
            show_plot = False
            df = prepare(temp['L2D'], -2, limit_N=limit_N, simple=True)
        else:
            N, df = prepare(temp['L2D'], -2, limit_N=limit_N, simple=False)

        if self.verbose:
            print(f"\nTest performed on {len(temp)} registries.\n\nDiscarded "
                  f"{len(self) - len(temp)} records < 1000 after preparation")
            if confidence is not None:
                _inform_(df, high_Z, conf)

        # Mean absolute difference
        if MAD:
            self.MAD = df.AbsDif.mean()
            if self.verbose:
                _report_mad_(-2, self.MAD)
        # Mean Square Error
        if MSE:
            self.MSE = (df.AbsDif ** 2).mean()

        # Chi-square statistic
        if chi_square:
            self.chi_square = chi_sq(df, ddf=99, confidence=confidence,
                                     verbose=self.verbose)
        # KS test
        if KS:
            self.KS = kolmogorov_smirnov(df, confidence=confidence, N=len(temp),
                                         verbose=self.verbose)

        if bhat_coeff:
            self.bhat_coeff = _bhattacharyya_coefficient(
                                df.Found.values, df.Expected.values)

        if bhat_dist:
            self.bhat_dist = _bhattacharyya_distance_(
                                df.Found.values, df.Expected.values)
        
        if kl_diverg:
            self.kl_diverg = _kullback_leibler_divergence_(
                                df.Found.values, df.Expected.values)

        # Plotting expected frequencies (line) versus found ones (bars)
        if show_plot:
            plot_digs(df, x=arange(0, 100), y_Exp=df.Expected,
                       y_Found=df.Found, N=N, figsize=(15, 5),
                       conf_Z=conf, text_x=True, save_plot=save_plot,
                       save_plot_kwargs=save_plot_kwargs)
        if ret_df:
            return df

    def summation(self, digs=2, top=20, show_plot=True, save_plot=None,
                  save_plot_kwargs=None, ret_df=False):
        """Performs the Summation test. In a Benford series, the sums of the
        entries begining with the same digits tends to be the same.

        Args:
            digs: tells the first digits to use. 1- first; 2- first two;
                3- first three. Defaults to 2.
            top: choses how many top values to show. Defaults to 20.
            show_plot: plots the results. Defaults to True.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        
        Returns:
            DataFrame with the Expected and Found proportions, and their
                absolute differences
        """
        _check_digs_(digs)

        if digs == 1:
            top = 9
        # Call the dict for F1D, F2D, F3D
        d = digs_dict[digs]
        if d not in self.columns:
            self[d] = self.ZN.astype(str).str[:digs].astype(int)
        # Call the expected proportion according to digs
        li = 1. / (9 * (10 ** (digs - 1)))

        df = self.groupby(d).sum()
        # s.drop(0, inplace=True)
        df['Percent'] = df.ZN / df.ZN.sum()
        df.columns.values[1] = 'Summ'
        df = df[['Summ', 'Percent']]
        df['AbsDif'] = (df.Percent - li).abs()

        if self.verbose:
            # N = len(self)
            print(f"\nTest performed on {len(self)} registries.\n")
            print(f"The top {top} diferences are:\n")
            print(df[:top])

        if show_plot:
            plot_sum(df, figsize=(
                       2 * (digs ** 2 + 5), 1.5 * (digs ** 2 + 5)), li=li,
                       save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)

        if ret_df:
            return df

    def duplicates(self, top_Rep=20, inform=None):
        """Performs a duplicates test and maps the duplicates count in descending
        order.

        Args:
            verbose (bool): tells how many duplicated entries were found and prints the
                top numbers according to the top_Rep argument. Defaluts to True.
            top_Rep: int or None. Chooses how many duplicated entries will be
                shown withe the top repititions. Defaluts to 20. If None, returns
                al the ordered repetitions.

        Returns:
            DataFrame with the duplicated records and their occurrence counts,
                in descending order (if verbose is False; if True, prints to
                terminal).

        Raises:
            ValueError: if the `top_Rep` arg is not int or None.
        """
        if top_Rep is not None and not isinstance(top_Rep, int):
            raise ValueError('The top_Rep argument must be an int or None.')

        dup = self[['seq']][self.seq.duplicated(keep=False)]
        dup_count = dup.groupby(self.seq).count()

        dup_count.index.names = ['Entries']
        dup_count.rename(columns={'seq': 'Count'}, inplace=True)

        dup_count.sort_values('Count', ascending=False, inplace=True)

        # self.maps['dup'] = dup_count.index[:top_Rep].values  # array

        if self.verbose:
            print(f'\nFound {len(dup_count)} duplicated entries.\n'
                  f'The entries with the {top_Rep} highest repitition counts are:')
            print(dup_count.head(top_Rep))
        else:
            return dup_count


class Mantissas(object):
    """
    Returns a Series with the data mantissas,

    Args:
        data: sequence to compute mantissas from, numpy 1D array, pandas
            Series of pandas DataFrame column.
    Attributes:
        data (DataFrame): holds the computed mantissas and, if the arc_test
            is also called, the respecttive x and Y coordinates for the plot.
        stats (dict): holds the relevant statistics about the data mantissas.
    """

    def __init__(self, data):

        data = Series(_check_num_array_(data))
        data = data.dropna().loc[data != 0].abs()

        self.data = DataFrame({'Mantissa': get_mantissas(data.abs())})

        self.stats = {'Mean': self.data.Mantissa.mean(),
                      'Var': self.data.Mantissa.var(),
                      'Skew': self.data.Mantissa.skew(),
                      'Kurt': self.data.Mantissa.kurt()}

    def report(self, show_plot=True, save_plot=None, save_plot_kwargs=None):
        """Displays the Mantissas stats.

        Args:
            show_plot: shows the ordered mantissas plot and the Arc Test plot.
                Defaults to True.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension. Only available when
                plot=True.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when plot=True and save_plot is a string with the
                figure file path/name.
        """
        print("\n", '  Mantissas Test  '.center(52, '#'))
        print(f"\nThe Mantissas MEAN is      {self.stats['Mean']:.6f}."
              "\tRef: 0.5")
        print(f"The Mantissas VARIANCE is  {self.stats['Var']:.6f}."
              "\tRef: 0.08333")
        print(f"The Mantissas SKEWNESS is  {self.stats['Skew']:.6f}."
              "\tRef: 0.0")
        print(f"The Mantissas KURTOSIS is  {self.stats['Kurt']:.6f}."
              "\tRef: -1.2\n")
        if show_plot:
            self.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
            self.arc_test(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)

    def show_plot(self, figsize=(12, 12), save_plot=None, save_plot_kwargs=None):
        """Plots the ordered mantissas and compares them to the expected, straight
        line that should be formed in a Benford-cmpliant set.

        Args:
            figsize: tuple that sets the figure size.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        plot_ordered_mantissas(self.data.Mantissa, figsize=figsize,
                               save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
 
    def arc_test(self, grid=True, figsize=12, save_plot=None, save_plot_kwargs=None):
        """
        Add two columns to Mantissas's DataFrame equal to their "X" and "Y"
        coordinates, plots its to a scatter plot and calculates the gravity
        center of the circle.

        Args:
            grid:show grid of the plot. Defaluts to True.
            figsize: size of the figure to be displayed. Since it is a square,
                there is no need to provide a tuple, like is usually the case with
                matplotlib.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        if self.stats.get('gravity_center') is None:
            self.data['mant_x'] = cos(2 * pi * self.data.Mantissa)
            self.data['mant_y'] = sin(2 * pi * self.data.Mantissa)
            self.stats['gravity_center'] = (self.data.mant_x.mean(),
                                            self.data.mant_y.mean())
        plot_mantissa_arc_test(self.data, self.stats['gravity_center'],
                               figsize=figsize, save_plot=save_plot,
                               save_plot_kwargs=save_plot_kwargs)


class Roll_mad(object):
    """Applies the MAD to sequential subsets of the Series, returning another
    Series.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: tells which test to use. 1: Fisrt Digits; 2: First Two Digits;
            3: First Three Digits; 22: Second Digit; and -2: Last Two Digits.
        window: size of the subset to be used.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.

    """

    def __init__(self, data, test, window, decimals=2, sign='all'):

        #: the test (F1D, SD, F2D...) used for the MAD calculation and critical values
        self.test = _check_test_(test)

        if not isinstance(data, Source):
            data = Source(data, sign=sign, decimals=decimals, verbose=False)

        Exp, ind = prep_to_roll(data, self.test)

        self.roll_series = data[digs_dict[test]].rolling(
                                window=window).apply(mad_to_roll, 
                                    args=(Exp, ind), raw=False)
        self.roll_series.dropna(inplace=True)

    def show_plot(self, figsize=(15, 8), save_plot=None, save_plot_kwargs=None):
        """Shows the rolling MAD plot

        Args:
            figsize: the figure dimensions.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        plot_roll_mad(self, figsize=figsize,
                      save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)


class Roll_mse(object):
    """Applies the MSE to sequential subsets of the Series, returning another
    Series.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: tells which test to use. 1: Fisrt Digits; 2: First Two Digits;
            3: First Three Digits; 22: Second Digit; and -2: Last Two Digits.
        window: size of the subset to be used.
            decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. 'pos': only the positive
            entries; 'neg': only negative entries; 'all': all entries but zeros.
            Defaults to 'all'.
    """

    def __init__(self, data, test, window, decimals=2, sign='all'):

        test = _check_test_(test)

        if not isinstance(data, Source):
            data = Source(data, sign=sign, decimals=decimals, verbose=False)

        Exp, ind = prep_to_roll(data, test)

        self.roll_series = data[digs_dict[test]].rolling(
                                window=window).apply(mse_to_roll, 
                                    args=(Exp, ind), raw=False)
        self.roll_series.dropna(inplace=True)

    def show_plot(self, figsize=(15, 8), save_plot=None, save_plot_kwargs=None):
        """Shows the rolling MSE plot

        Args:
            figsize: the figure dimensions.
            save_plot (str): string with the path/name of the file in which the generated
                plot will be saved. Uses matplotlib.pyplot.savefig(). File format
                is infered by the file name extension.
            save_plot_kwargs (dict): any of the kwargs accepted by
                matplotlib.pyplot.savefig()
                https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
                Only available when save_plot is a string with the figure file
                path/name.
        """
        plot_roll_mse(self.roll_series, figsize=figsize,
                      save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)


def first_digits(data, digs, decimals=2, sign='all', verbose=True,
                 confidence=None, high_Z='pos', limit_N=None,
                 MAD=False, MSE=False, chi_square=False, KS=False,
                 show_plot=True, save_plot=None, save_plot_kwargs=None,
                 inform=None):
    """Performs the Benford First Digits test on the series of
    numbers provided.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. 'pos': only the positive
            entries; 'neg': only negative entries; 'all': all entries but zeros.
            Defaults to 'all'.
        digs (int): number of first digits to consider. Must be 1 (first digit),
            2 (first two digits) or 3 (first three digits).
        verbose (bool): tells the number of registries that are being subjected to
            the analysis and returns tha analysis DataFrame sorted by the
            highest Z score down. Defaults to True.
        MAD (bool): calculates the Mean Absolute Difference between the
            found and the expected distributions; defaults to False.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show. Defaults to None.
        high_Z (int): chooses which Z scores to be used when displaying results,
            according to the confidence level chosen. Defaluts to 'pos',
            which will highlight only values higher than the expexted
            frequencies; 'all' will highlight both extremes (positive and
            negative); and an integer, which will use the first n entries,
            positive and negative, regardless of whether Z is higher than
            the confidence or not.
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.
        MSE (bool): calculates the Mean Square Error of the sample; defaults to
            False.
        chi_square: calculates the chi_square statistic of the sample and
            compares it with a critical value, according to the confidence
            level chosen and the series's degrees of freedom. Defaults to
            False. Requires confidence != None.
        KS: calculates the Kolmogorov-Smirnov test, comparing the cumulative
            distribution of the sample with the Benford's, according to the
            confidence level chosen. Defaults to False. Requires confidence
            != None.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        DataFrame with the Expected and Found proportions, and the Z scores of
            the differences if the confidence is not None.
    """
    verbose = _deprecate_inform_(verbose, inform)

    if not isinstance(data, Source):
        data = Source(data, decimals=decimals, sign=sign, verbose=verbose)

    data = data.first_digits(digs, confidence=confidence, high_Z=high_Z,
                             limit_N=limit_N, MAD=MAD, MSE=MSE,
                             chi_square=chi_square, KS=KS, show_plot=show_plot,
                             save_plot=save_plot, save_plot_kwargs=save_plot_kwargs,
                             ret_df=True)

    if confidence is not None:
        data = data[['Counts', 'Found', 'Expected', 'Z_score']]
        return data.sort_values('Z_score', ascending=False)
    else:
        return data[['Counts', 'Found', 'Expected']]


def second_digit(data, decimals=2, sign='all', verbose=True,
                 confidence=None, high_Z='pos', limit_N=None,
                 MAD=False, MSE=False, chi_square=False, KS=False,
                 show_plot=True, save_plot=None, save_plot_kwargs=None,
                 inform=None):
    """Performs the Benford Second Digits test on the series of
    numbers provided.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. 'pos': only the positive
            entries; 'neg': only negative entries; 'all': all entries but zeros.
            Defaults to 'all'.
        verbose (bool): tells the number of registries that are being subjected to
            the analysis and returns tha analysis DataFrame sorted by the
            highest Z score down. Defaults to True.
        MAD (bool): calculates the Mean Absolute Difference between the
            found and the expected distributions; defaults to False.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show. Defaults to None.
        high_Z (int): chooses which Z scores to be used when displaying results,
            according to the confidence level chosen. Defaluts to 'pos',
            which will highlight only values higher than the expexted
            frequencies; 'all' will highlight both extremes (positive and
            negative); and an integer, which will use the first n entries,
            positive and negative, regardless of whether Z is higher than
            the confidence or not.
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.
        MSE (bool): calculates the Mean Square Error of the sample; defaults to
            False.
        chi_square: calculates the chi_square statistic of the sample and
            compares it with a critical value, according to the confidence
            level chosen and the series's degrees of freedom. Defaults to
            False. Requires confidence != None.
        KS: calculates the Kolmogorov-Smirnov test, comparing the cumulative
            distribution of the sample with the Benford's, according to the
            confidence level chosen. Defaults to False. Requires confidence
            != None.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.

    Returns:
        DataFrame with the Expected and Found proportions, and the Z scores of
            the differences if the confidence is not None.
    """
    verbose = _deprecate_inform_(verbose, inform)

    if not isinstance(data, Source):
        data = Source(data, sign=sign, decimals=decimals, verbose=verbose)

    data = data.second_digit(confidence=confidence, high_Z=high_Z,
                             limit_N=limit_N, MAD=MAD, MSE=MSE,
                             chi_square=chi_square, KS=KS, show_plot=show_plot,
                             save_plot=save_plot, save_plot_kwargs=save_plot_kwargs,
                             ret_df=True)
    if confidence is not None:
        data = data[['Counts', 'Found', 'Expected', 'Z_score']]
        return data.sort_values('Z_score', ascending=False)
    else:
        return data[['Counts', 'Found', 'Expected']]


def last_two_digits(data, decimals=2, sign='all', verbose=True,
                    confidence=None, high_Z='pos', limit_N=None,
                    MAD=False, MSE=False, chi_square=False, KS=False,
                    show_plot=True, save_plot=None, save_plot_kwargs=None,
                    inform=None):
    """Performs the Last Two Digits test on the series of
    numbers provided.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column,with values being
            integers or floats.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. 'pos': only the positive
            entries; 'neg': only negative entries; 'all': all entries but zeros.
            Defaults to 'all'.
        verbose (bool): tells the number of registries that are being subjected to
            the analysis and returns tha analysis DataFrame sorted by the
            highest Z score down. Defaults to True.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show. Defaults to None.
        high_Z (int): chooses which Z scores to be used when displaying results,
            according to the confidence level chosen. Defaluts to 'pos',
            which will highlight only values higher than the expexted
            frequencies; 'all' will highlight both extremes (positive and
            negative); and an integer, which will use the first n entries,
            positive and negative, regardless of whether Z is higher than
            the confidence or not.
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.
        MAD (bool): calculates the Mean Absolute Difference between the
            found and the expected distributions; defaults to False.
        MSE (bool): calculates the Mean Square Error of the sample; defaults to
            False.
        chi_square: calculates the chi_square statistic of the sample and
            compares it with a critical value, according to the confidence
            level chosen and the series's degrees of freedom. Defaults to
            False. Requires confidence != None.
        KS: calculates the Kolmogorov-Smirnov test, comparing the cumulative
            distribution of the sample with the Benford's, according to the
            confidence level chosen. Defaults to False. Requires confidence
            != None.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.

    Returns:
        DataFrame with the Expected and Found proportions, and the Z scores of
            the differences if the confidence is not None.
    """
    verbose = _deprecate_inform_(verbose, inform)

    if not isinstance(data, Source):
        data = Source(data, decimals=decimals, sign=sign, verbose=verbose)

    data = data.last_two_digits(confidence=confidence, high_Z=high_Z,
                                limit_N=limit_N, MAD=MAD,
                                MSE=MSE, chi_square=chi_square, KS=KS,
                                show_plot=show_plot, save_plot=save_plot,
                                save_plot_kwargs=save_plot_kwargs, ret_df=True)

    if confidence is not None:
        data = data[['Counts', 'Found', 'Expected', 'Z_score']]
        return data.sort_values('Z_score', ascending=False)
    else:
        return data[['Counts', 'Found', 'Expected']]


def mantissas(data, report=True, show_plot=True, arc_test=True,
              save_plot=None, save_plot_kwargs=None, inform=None):
    """Extraxts the mantissas of the records logarithms

    Args:
        data: sequence to compute mantissas from, numpy 1D array, pandas Series
            of pandas DataFrame column.
        report: prints the mamtissas mean, variance, skewness and kurtosis
            for the sequence studied, along with reference values.
        show_plot: plots the ordered mantissas and a line with the expected
            inclination. Defaults to True.
        arc_test: draws the Arc Test plot. Defaluts to True.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        Series with the data mantissas.
    """
    report = _deprecate_inform_(report, inform)

    mant = Mantissas(data)
    if report:
        mant.report(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    if show_plot:
        mant.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    if arc_test:
        mant.arc_test(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    return mant


def summation(data, digs=2, decimals=2, sign='all', top=20, verbose=True,
              show_plot=True, save_plot=None, save_plot_kwargs=None, inform=None):
    """Performs the Summation test. In a Benford series, the sums of the
    entries begining with the same digits tends to be the same.
    Works only with the First Digits (1, 2 or 3) test.

    Args:
        digs: tells the first digits to use: 1- first; 2- first two;
            3- first three. Defaults to 2.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        top: choses how many top values to show. Defaults to 20.
        show_plot: plots the results. Defaults to True.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        DataFrame with the Summation test, whether sorted in descending order
            (if verbose == True) or not.
    """
    verbose = _deprecate_inform_(verbose, inform)

    if not isinstance(data, Source):
        data = Source(data, sign=sign, decimals=decimals, verbose=verbose)

    data = data.summation(digs=digs, top=top,
                          show_plot=show_plot, save_plot=save_plot,
                          save_plot_kwargs=save_plot_kwargs, ret_df=True)
    if verbose:
        return data.sort_values('AbsDif', ascending=False)
    else:
        return data


def mad(data, test, decimals=2, sign='all', verbose=False):
    """Calculates the Mean Absolute Deviation of the Series

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: informs which base test to use for the mad.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.

    Returns:
        float: the Mean Absolute Deviation of the Series
    """
    data = _check_num_array_(data)
    test = _check_test_(test)
    start = Source(data, sign=sign, decimals=decimals, verbose=verbose)
    if test in [1, 2, 3]:
        start.first_digits(digs=test, MAD=True, MSE=True, simple=True)
    elif test == 22:
        start.second_digit(MAD=True, MSE=False, simple=True)
    else:
        start.last_two_digits(MAD=True, MSE=False, simple=True)
    return start.MAD


def mse(data, test, decimals=2, sign='all', verbose=False):
    """Calculates the Mean Squared Error of the Series

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: informs which base test to use for the mad.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.

    Returns:
        float: the Mean Squared Error of the Series
    """
    data = _check_num_array_(data)
    test = _check_test_(test)
    start = Source(data, sign=sign, decimals=decimals, verbose=verbose)
    if test in [1, 2, 3]:
        start.first_digits(digs=test, MAD=False, MSE=True, simple=True)
    elif test == 22:
        start.second_digit(MAD=False, MSE=True, simple=True)
    else:
        start.last_two_digits(MAD=False, MSE=True, simple=True)
    return start.MSE


def bhattacharyya_distance(data, test, decimals, sign="all", verbose=False):
    """Computes the Bhattacharyya Distance between the Found and the Expected
    (Benford) digits distributions, according toe the test chosen
    (First, Second, First Two...)

    Args:
        data (ndarray, Series): sequence to be evaluated, with values being
            integers or floats.
        test (int, str): informs which base test to be used.
        decimals (int): number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign (str, optional): tells which portion of the data to consider.
            pos: only the positive entries; neg: only negative entries; all:
            all entries but zeros. Defaults to "all".

    Returns:
        float: the Bhattacharyya Distance between the distributions
    """
    data = _check_num_array_(data)
    test = _check_test_(test)
    start = Source(data, sign=sign, decimals=decimals, verbose=verbose)
    if test in [1, 2, 3]:
        start.first_digits(digs=test, MAD=False, bhat_dist=True, simple=True)
    elif test == 22:
        start.second_digit(MAD=False, bhat_dist=True, simple=True)
    else:
        start.last_two_digits(MAD=False, bhat_dist=True, simple=True)
    return start.bhat_dist


def kullback_leibler_divergence(data, test, decimals, sign="all",
                                verbose=False):
    """Computes the Kulback-Leibler Divergence between the Found and the
    Expected (Benford) digits distributions, according toe the test chosen
    (First, Second, First Two...).

    Args:
        data (ndarray, Series): sequence to be evaluated, with values being
            integers or floats.
        test (int, str): informs which base test to be used.
        decimals (int): number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign (str, optional): tells which portion of the data to consider.
            pos: only the positive entries; neg: only negative entries; all:
            all entries but zeros. Defaults to "all".

    Returns:
        float: the Kulback-Leibler Divergence between the distributions
    """
    data = _check_num_array_(data)
    test = _check_test_(test)
    start = Source(data, sign=sign, decimals=decimals, verbose=verbose)
    if test in [1, 2, 3]:
        start.first_digits(digs=test, MAD=False, kl_diverg=True, simple=True)
    elif test == 22:
        start.second_digit(MAD=False, kl_diverg=True, simple=True)
    else:
        start.last_two_digits(MAD=False, kl_diverg=True, simple=True)
    return start.kl_diverg


def mad_summ(data, test, decimals=2, sign='all', verbose=False):
    """Calculate the Mean Absolute Deviation of the Summation Test

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: informs which base test to use for the summation mad.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.

    Returns:
        float: the Mean Absolute Deviation of the Summation Test
    """
    data = _check_num_array_(data)
    test = _check_digs_(test)

    start = Source(data, sign=sign, decimals=decimals, verbose=verbose)
    temp = start.loc[start.ZN >= 10 ** (test - 1)]
    temp[digs_dict[test]] = (temp.ZN // 10 ** ((log10(temp.ZN).astype(
                                                int)) - (test - 1))).astype(
                                                    int)
    li = 1. / (9 * (10 ** (test - 1)))

    df = temp.groupby(digs_dict[test]).sum()
    return mean(abs(df.ZN / df.ZN.sum() - li))


def rolling_mad(data, test, window, decimals=2, sign='all',
                show_plot=False, save_plot=None, save_plot_kwargs=None):
    """Applies the MAD to sequential subsets of the records.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: tells which test to use. 1: Fisrt Digits; 2: First Two Digits;
            3: First Three Digits; 22: Second Digit; and -2: Last Two Digits.
        window: size of the subset to be used.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        Series with sequentially computed MADs.
    """
    data = _check_num_array_(data)
    r_mad = Roll_mad(data, test, window, decimals, sign)
    if show_plot:
        r_mad.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    return r_mad.roll_series


def rolling_mse(data, test, window, decimals=2, sign='all',
                show_plot=False, save_plot=None, save_plot_kwargs=None):
    """Applies the MSE to sequential subsets of the records.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: tells which test to use. 1: Fisrt Digits; 2: First Two Digits;
            3: First Three Digits; 22: Second Digit; and -2: Last Two Digits.
        window: size of the subset to be used.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        Series with sequentially computed MSEs.
    """
    data = _check_num_array_(data)
    r_mse = Roll_mse(data, test, window, decimals, sign)
    if show_plot:
        r_mse.show_plot(save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    return r_mse.roll_series


def duplicates(data, top_Rep=20, verbose=True, inform=None):
    """Performs a duplicates test and maps the duplicates count in descending
    order.

    Args:
        data: sequence to take the duplicates from. pandas Series or
            numpy Ndarray.
        verbose (bool): tells how many duplicated entries were found and prints the
            top numbers according to the top_Rep argument. Defaluts to True.
        top_Rep: chooses how many duplicated entries will be
            shown withe the top repititions. int or None. Defaluts to 20.
            If None, returns al the ordered repetitions.

    Returns:
        DataFrame with the duplicated records and their respective counts

    Raises:
        ValueError: if the `top_Rep` arg is not int or None.
    """
    verbose = _deprecate_inform_(verbose, inform)

    if top_Rep is not None and not isinstance(top_Rep, int):
        raise ValueError('The top_Rep argument must be an int or None.')

    if not isinstance(data, Series):
        try:
            data = Series(data)
        except ValueError:
            print('\ndata must be a numpy Ndarray or a pandas Series.')

    dup = data.loc[data.duplicated(keep=False)]
    dup_count = dup.value_counts()

    dup_count.index.names = ['Entries']
    dup_count.name = 'Count'

    if verbose:
        print(f'\nFound {len(dup_count)} duplicated entries.\n'
              f'The entries with the {top_Rep} highest repitition counts are:')
        print(dup_count.head(top_Rep))

    return dup_count


def second_order(data, test, decimals=2, sign='all', verbose=True, MAD=False,
                 confidence=None, high_Z='pos', limit_N=None, MSE=False,
                 show_plot=True, save_plot=None, save_plot_kwargs=None, inform=None):
    """Performs the chosen test after subtracting the ordered sequence by itself.
    Hence Second Order.

    Args:
        data: sequence of numbers to be evaluated. Must be a numpy 1D array,
            a pandas Series or a pandas DataFrame column, with values being
            integers or floats.
        test: the test to be performed - 1 or 'F1D': First Digit; 2 or 'F2D':
            First Two Digits; 3 or 'F3D': First three Digits; 22 or 'SD':
            Second Digits; -2 or 'L2D': Last Two Digits.
        decimals: number of decimal places to consider. Defaluts to 2.
            If integers, set to 0. If set to -infer-, it will remove the zeros
            and consider up to the fifth decimal place to the right, but will
            loose performance.
        sign: tells which portion of the data to consider. pos: only the positive
            entries; neg: only negative entries; all: all entries but zeros.
            Defaults to all.
        verbose (bool): tells the number of registries that are being subjected to
            the analysis and returns tha analysis DataFrame sorted by the
            highest Z score down. Defaults to True.
        MAD (bool): calculates the Mean Absolute Difference between the
            found and the expected distributions; defaults to False.
        confidence (int, float): confidence level to draw lower and upper limits when
            plotting and to limit the top deviations to show. Defaults to None.
        high_Z (int): chooses which Z scores to be used when displaying results,
            according to the confidence level chosen. Defaluts to 'pos',
            which will highlight only values higher than the expexted
            frequencies; 'all' will highlight both extremes (positive and
            negative); and an integer, which will use the first n entries,
            positive and negative, regardless of whether Z is higher than
            the confidence or not.
        limit_N (int): sets a limit to N as the sample size for the calculation of
            the Z scores if the sample is too big. Defaults to None.
        MSE (bool): calculates the Mean Square Error of the sample; defaults to
            False.
        chi_square: calculates the chi_square statistic of the sample and
            compares it with a critical value, according to the confidence
            level chosen and the series's degrees of freedom. Defaults to
            False. Requires confidence != None.
        KS: calculates the Kolmogorov-Smirnov test, comparing the cumulative
            distribution of the sample with the Benford's, according to the
            confidence level chosen. Defaults to False. Requires confidence
            != None.
        show_plot (bool): draws the test plot.
        save_plot (str): string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs (dict): any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    
    Returns:
        DataFrame of the test chosen, but applied on Second Order pre-
            processed data.
    """
    test = _check_test_(test)

    verbose = _deprecate_inform_(verbose, inform)

    data = Source(data, decimals=decimals, sign=sign,
                  sec_order=True, verbose=verbose)
    if test in [1, 2, 3]:
        data.first_digits(digs=test, MAD=MAD,
                          confidence=confidence, high_Z=high_Z,
                          limit_N=limit_N, MSE=MSE, show_plot=show_plot,
                          save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    elif test == 22:
        data.second_digit(MAD=MAD, confidence=confidence, high_Z=high_Z,
                          limit_N=limit_N, MSE=MSE, show_plot=show_plot,
                          save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    else:
        data.last_two_digits(MAD=MAD, confidence=confidence, high_Z=high_Z,
                             limit_N=limit_N, MSE=MSE, show_plot=show_plot,
                             save_plot=save_plot, save_plot_kwargs=save_plot_kwargs)
    return data



from numpy import sqrt, log, where
from .constants import crit_chi2, KS_crit, mad_dict, digs_dict


def Z_score(frame, N):
    """Computes the Z statistics for the proportions studied

    Args:
        frame: DataFrame with the expected proportions and the already calculated
            Absolute Diferences between the found and expeccted proportions
        N: sample size

    Returns:
        Series of computed Z scores
    """
    return (frame.AbsDif - (1 / (2 * N))) / sqrt(
           (frame.Expected * (1. - frame.Expected)) / N)


def chi_sq(frame, ddf, confidence, verbose=True):
    """Comnputes the chi-square statistic of the found distributions and compares
    it with the critical chi-square of such a sample, according to the
    confidence level chosen and the degrees of freedom - len(sample) -1.

    Args:
        frame: DataFrame with Found, Expected and their difference columns.
        ddf: Degrees of freedom to consider.
        confidence: Confidence level to look up critical value.
        verbose: prints the chi-squre result and compares to the critical
            chi-square for the sample. Defaults to True.

    Returns:
        The computed Chi square statistic and the critical chi square
            (according) to the degrees of freedom and confidence level,
            for comparison. None if confidence is None
    """
    if confidence is None:
        print('\nChi-square test needs confidence other than None.')
        return
    else:
        exp_counts = frame.Counts.sum() * frame.Expected
        dif_counts = frame.Counts - exp_counts
        found_chi = (dif_counts ** 2 / exp_counts).sum()
        crit_chi = crit_chi2[ddf][confidence]
        if verbose:
            print(f"\nThe Chi-square statistic is {found_chi:.4f}.\n"
                  f"Critical Chi-square for this series: {crit_chi}.")
        return (found_chi, crit_chi)


def chi_sq_2(frame):
    """Computes the chi-square statistic of the found distributions

    Args:
        frame: DataFrame with Found, Expected and their difference columns.

    Returns:
        The computed Chi square statistic 
    """
    exp_counts = frame.Counts.sum() * frame.Expected
    dif_counts = frame.Counts - exp_counts
    return (dif_counts ** 2 / exp_counts).sum()


def kolmogorov_smirnov(frame, confidence, N, verbose=True):
    """Computes the Kolmogorov-Smirnov test of the found distributions
    and compares it with the critical chi-square of such a sample,
    according to the confidence level chosen.

    Args:
        frame: DataFrame with Foud and Expected distributions.
        confidence: Confidence level to look up critical value.
        N: Sample size
        verbose: prints the KS result and the critical value for the sample.
            Defaults to True.

    Returns:
        The Suprem, which is the greatest absolute difference between the
            Found and the expected proportions, and the Kolmogorov-Smirnov
            critical value according to the confidence level, for ccomparison
    """
    if confidence is None:
        print('\nKolmogorov-Smirnov test needs confidence other than None.')
        return
    else:
        # sorting and calculating the cumulative distribution
        ks_frame = frame.sort_index()[['Found', 'Expected']].cumsum()
        # finding the supremum - the largest cumul dist difference
        suprem = ((ks_frame.Found - ks_frame.Expected).abs()).max()
        # calculating the crittical value according to confidence
        crit_KS = KS_crit[confidence] / sqrt(N)

        if verbose:
            print(f"\nThe Kolmogorov-Smirnov statistic is {suprem:.4f}.\n"
                  f"Critical K-S for this series: {crit_KS:.4f}")
        return (suprem, crit_KS)


def kolmogorov_smirnov_2(frame):
    """Computes the Kolmogorov-Smirnov test of the found distributions

    Args:
        frame: DataFrame with Foud and Expected distributions.

    Returns:
        The Suprem, which is the greatest absolute difference between the
            Found end th expected proportions
    """
    # sorting and calculating the cumulative distribution
    ks_frame = frame.sort_index()[['Found', 'Expected']].cumsum()
    # finding the supremum - the largest cumul dist difference
    return ((ks_frame.Found - ks_frame.Expected).abs()).max()


def mad(frame, test, verbose=True):
    """Computes the Mean Absolute Deviation (MAD) between the found and the
    expected proportions.

    Args:
        frame: DataFrame with the Absolute Deviations already calculated.
        test: Test to compute the MAD from (F1D, SD, F2D...)
        verbose: prints the MAD result and compares to limit values of
            conformity. Defaults to True.

    Returns:
        The Mean of the Absolute Deviations between the found and expected
            proportions. 
    """
    mad = frame.AbsDif.mean()

    if verbose:
        print(f"\nThe Mean Absolute Deviation is {mad}")

        if test != -2:
            print(f"For the {mad_dict[digs_dict[test]]}:\n\
            - 0.0000 to {mad_dict[test][0]}: Close Conformity\n\
            - {mad_dict[test][0]} to {mad_dict[test][1]}: Acceptable Conformity\n\
            - {mad_dict[test][1]} to {mad_dict[test][2]}: Marginally Acceptable Conformity\n\
            - Above {mad_dict[test][2]}: Nonconformity")
        else:
            pass
    return mad


def mse(frame, verbose=True):
    """Computes the test's Mean Square Error

    Args:
        frame: DataFrame with the already computed Absolute Deviations between
            the found and expected proportions
        verbose: Prints the MSE. Defaults to True.

    Returns:
        Mean of the squared differences between the found and the expected proportions.
    """
    mse = (frame.AbsDif ** 2).mean()

    if verbose:
        print(f"\nMean Square Error = {mse}")

    return mse

def _bhattacharyya_coefficient(dist_1, dist_2):
    """Computes the Bhattacharyya Coeficient between two probability
    distributions, to be letar used to compute the Bhattacharyya Distance

    Args:
        dist_1 (np.array): The newly gathered distribution, to be compared
            with an older / established distribution.
        dist_2 (np.array): The older/ establhished distribution with which
            the new one will be compared. 
    
    Returns:
        bhat_coef (float)
    """
    return sqrt(dist_1 * dist_2).sum()


def _bhattacharyya_distance_(dist_1, dist_2):
    """Computes the Bhattacharyya Dsitance between two probability
    distributions

    Args:
        dist_1 (np.array): The newly gathered distribution, to be compared
            with an older / established distribution.
        dist_2 (np.array): The older/ establhished distribution with which
            the new one will be compared. 
    
    Returns:
        bhat_dist (float)
    """
    return -log(_bhattacharyya_coefficient(dist_1, dist_2))


def _kullback_leibler_divergence_(dist_1, dist_2):
    """Computes the Kullback-Leibler Divergence between two probability
    distributions.

    Args:
        dist_1 (np.array): The newly gathered distribution, to be compared
            with an older / established distribution.
        dist_2 (np.array): The older/ establhished distribution with which
            the new one will be compared. 

    Returns:
        kulb_leib_diverg (float)        
    """
    return (log((dist_1 / dist_2), where=(dist_1 != 0)) * dist_1).sum()


from pandas import Series
from numpy import array, ndarray
from .constants import digs_dict, rev_digs, confs


def _check_digs_(digs):
    """Checks the possible values for the digs parameter of the
    First Digits tests
    """
    if digs not in [1, 2, 3]:
        raise ValueError("The value assigned to the parameter -digs- "
                         f"was {digs}. Value must be 1, 2 or 3.")


def _check_test_(test):
    """Checks the test chosen, both for int or str values
    """
    if isinstance(test, int):
        if test in digs_dict.keys():
            return test
        else:
            raise ValueError(f'Test was set to {test}. Should be one of '
                             f'{digs_dict.keys()}')
    elif isinstance(test, str):
        if test in rev_digs.keys():
            return rev_digs[test]
        else:
            raise ValueError(f'Test was set to {test}. Should be one of '
                             f'{rev_digs.keys()}')
    else:
        raise ValueError('Wrong value chosen for test parameter. Possible '
                         f'values are\n {list(digs_dict.keys())} for ints and'
                         f'\n {list(rev_digs.keys())} for strings.')


def _check_decimals_(decimals):
    """"""
    if isinstance(decimals, int):
        if (decimals < 0):
            raise ValueError(
                "Parameter -decimals- must be an int >= 0, or 'infer'.")
    else:
        if decimals != 'infer':
            raise ValueError(
                "Parameter -decimals- must be an int >= 0, or 'infer'.")
    return decimals


def _check_sign_(sign):
    """"""
    if sign not in ['all', 'pos', 'neg']:
        raise ValueError("Parameter -sign- must be one of the following: "
                         "'all', 'pos' or 'neg'.")
    return sign


def _check_confidence_(confidence):
    """"""
    if confidence not in confs.keys():
        raise ValueError("Value of parameter -confidence- must be one of the "
                         f"following:\n {list(confs.keys())}")
    return confidence


def _check_high_Z_(high_Z):
    """"""
    if not high_Z in ['pos', 'all']:
        if not isinstance(high_Z, int):
            raise ValueError("The parameter -high_Z- should be 'pos', "
                             "'all' or an int.")
    return high_Z


def _check_num_array_(data):
    """"""
    if (not isinstance(data, ndarray)) & (not isinstance(data, Series)):
        print('\n`data` not a numpy NDarray nor a pandas Series.'
              ' Trying to convert...')
        try:
            data = array(data)
        except:
            raise ValueError('Could not convert data. Check input.')
        print('\nConversion successful.')

        try:
            data = data.astype(float)
        except:
            raise ValueError('Could not convert data. Check input.')
    else:
        if data.dtype not in [int, float]:
            try:
                data = data.astype(float)
            except:
                raise ValueError('Could not convert data. Check input.')
    return data



from numpy import array, arange, maximum, sqrt, ones
import matplotlib.pyplot as plt
from matplotlib.text import Annotation
from .constants import colors, mad_dict


def plot_expected(df, digs, save_plot=None, save_plot_kwargs=None):
    """Plots the Expected Benford Distributions

    Args:
        df: DataFrame with the Expected Proportions
        digs: Test's digit
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    if digs in [1, 2, 3]:
        y_max = (df.Expected.max() + (10 ** -(digs) / 3)) * 100
        figsize = 2 * (digs ** 2 + 5), 1.5 * (digs ** 2 + 5)
    elif digs == 22:
        y_max = 13.
        figsize = 14, 10.5
    elif digs == -2:
        y_max = 1.1
        figsize = 15, 8
    fig, ax = plt.subplots(figsize=figsize)
    plt.title('Expected Benford Distributions', size='xx-large')
    plt.xlabel(df.index.name, size='x-large')
    plt.ylabel('Distribution (%)', size='x-large')
    ax.set_facecolor(colors['b'])
    ax.set_ylim(0, y_max)
    ax.bar(df.index, df.Expected * 100, color=colors['t'], align='center')
    ax.set_xticks(df.index)
    ax.set_xticklabels(df.index)

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False)


def _get_plot_args(digs):
    """Selects the correct arguments for the plotting functions, depending on the
    the test (digs) chosen.
    """
    if digs in [1, 2, 3]:
        text_x = False
        n, m = 10 ** (digs - 1), 10 ** (digs)
        x = arange(n, m)
        figsize = (2 * (digs ** 2 + 5), 1.5 * (digs ** 2 + 5))
    elif digs == 22:
        text_x = False
        x = arange(10)
        figsize = (14, 10)
    else:
        text_x = True
        x = arange(100)
        figsize = (15, 7)
    return x, figsize, text_x

def plot_digs(df, x, y_Exp, y_Found, N, figsize, conf_Z, text_x=False,
              save_plot=None, save_plot_kwargs=None):
    """Plots the digits tests results

    Args:
        df: DataFrame with the data to be plotted
        x: sequence to be used in the x axis
        y_Exp: sequence of the expected proportions to be used in the y axis
            (line)
        y_Found: sequence of the found proportions to be used in the y axis
            (bars)
        N: lenght of sequence, to be used when plotting the confidence levels
        figsize: tuple to state the size of the plot figure
        conf_Z: Confidence level
        save_pic: file path to save figure
        text_x: Forces to show all x ticks labels. Defaluts to True.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
        
    """
    if len(x) > 10:
        rotation = 90
    else:
        rotation = 0
    fig, ax = plt.subplots(figsize=figsize)
    plt.title('Expected vs. Found Distributions', size='xx-large')
    plt.xlabel('Digits', size='x-large')
    plt.ylabel('Distribution (%)', size='x-large')
    if conf_Z is not None:
        sig = conf_Z * sqrt(y_Exp * (1 - y_Exp) / N)
        upper = y_Exp + sig + (1 / (2 * N))
        lower_zeros = array([0]*len(upper))
        lower = maximum(y_Exp - sig - (1 / (2 * N)), lower_zeros)
        u = (y_Found < lower) | (y_Found > upper)
        c = array([colors['m']] * len(u))
        c[u] = colors['af']
        lower *= 100.
        upper *= 100.
        ax.plot(x, upper, color=colors['s'], zorder=5)
        ax.plot(x, lower, color=colors['s'], zorder=5)
        ax.fill_between(x, upper, lower, color=colors['s'],
                        alpha=.3, label='Conf')
    else:
        c = colors['m']
    ax.bar(x, y_Found * 100., color=c, label='Found', zorder=3, align='center')
    ax.plot(x, y_Exp * 100., color=colors['s'], linewidth=2.5,
            label='Benford', zorder=4)
    ax.set_xticks(x)
    ax.set_xticklabels(x, rotation=rotation)
    ax.set_facecolor(colors['b'])
    if text_x:
        ind = array(df.index).astype(str)
        ind[:10] = array(['00', '01', '02', '03', '04', '05',
                          '06', '07', '08', '09'])
        plt.xticks(x, ind, rotation='vertical')
    ax.legend()
    ax.set_ylim(0, max([y_Exp.max() * 100, y_Found.max() * 100]) + 10 / len(x))
    ax.set_xlim(x[0] - 1, x[-1] + 1)

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False)


def plot_sum(df, figsize, li, text_x=False, save_plot=None, save_plot_kwargs=None):
    """Plots the summation test results

    Args:
        df: DataFrame with the data to be plotted
        figsize: sets the dimensions of the plot figure
        li: value with which to draw the horizontal line
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    x = df.index
    rotation = 90 if len(x) > 10 else 0
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    plt.title('Expected vs. Found Sums')
    plt.xlabel('Digits')
    plt.ylabel('Sums')
    ax.bar(x, df.Percent, color=colors['m'],
           label='Found Sums', zorder=3, align='center')
    ax.set_xlim(x[0] - 1, x[-1] + 1)
    ax.axhline(li, color=colors['s'], linewidth=2, label='Expected', zorder=4)
    ax.set_xticks(x)
    ax.set_xticklabels(x, rotation=rotation)
    ax.set_facecolor(colors['b'])
    if text_x:
        ind = array(x).astype(str)
        ind[:10] = array(['00', '01', '02', '03', '04', '05',
                          '06', '07', '08', '09'])
        plt.xticks(x, ind, rotation='vertical')
    ax.legend()

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False)

def plot_ordered_mantissas(col, figsize=(12, 12),
                           save_plot=None, save_plot_kwargs=None):
    """Plots the ordered mantissas and compares them to the expected, straight
        line that should be formed in a Benford-cmpliant set.

    Args:
        col (Series): column of mantissas to plot.
        figsize (tuple): sets the dimensions of the plot figure.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
 
    """
    ld = len(col)
    x = arange(1, ld + 1)
    n = ones(ld) / ld
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111)
    ax.plot(x, col.sort_values(), linestyle='--',
            color=colors['s'], linewidth=3, label='Mantissas')
    ax.plot(x, n.cumsum(), color=colors['m'],
            linewidth=2, label='Expected')
    plt.ylim((0, 1.))
    plt.xlim((1, ld + 1))
    ax.set_facecolor(colors['b'])
    ax.set_title("Ordered Mantissas")
    plt.legend(loc='upper left')

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False);

def plot_mantissa_arc_test(df, gravity_center, grid=True, figsize=12,
                           save_plot=None, save_plot_kwargs=None):
    """Draws thee Mantissa Arc Test after computing X and Y circular coordinates
    for every mantissa and the center of gravity for the set

    Args:
        df (DataFrame): pandas DataFrame with the mantissas and the X and Y
            coordinates.
        gravity_center (tuple): coordinates for plottling the gravity center
        grid (bool): show grid. Defaults to True.
        figsize (int): figure dimensions. No need to be a tuple, since the
            figure is a square.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    fig = plt.figure(figsize=(figsize, figsize))
    ax = plt.subplot()
    ax.set_facecolor(colors['b'])
    ax.scatter(df.mant_x, df.mant_y, label="ARC TEST",
               color=colors['m'])
    ax.scatter(gravity_center[0], gravity_center[1],
               color=colors['s'])
    text_annotation = Annotation(
        "  Gravity Center: "
        f"x({round(gravity_center[0], 3)}),"
        f" y({round(gravity_center[1], 3)})",
        xy=(gravity_center[0] - 0.65,
            gravity_center[1] - 0.1),
        xycoords='data')
    ax.add_artist(text_annotation)
    ax.grid(True, which='both')
    ax.axhline(y=0, color='k')
    ax.axvline(x=0, color='k')
    ax.legend(loc='lower left')
    ax.set_title("Mantissas Arc Test")

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False);

def plot_roll_mse(roll_series, figsize, save_plot=None, save_plot_kwargs=None):
    """Shows the rolling MSE plot

    Args:
        roll_series: pd.Series resultant form rolling mse.
        figsize: the figure dimensions.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_facecolor(colors['b'])
    ax.plot(roll_series, color=colors['m'])

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False)

def plot_roll_mad(roll_mad, figsize, save_plot=None, save_plot_kwargs=None):
    """Shows the rolling MAD plot

    Args:
        roll_mad: pd.Series resultant form rolling mad.
        figsize: the figure dimensions.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
    """
    fig, ax = plt.subplots(figsize=figsize)
    ax.set_facecolor(colors['b'])
    ax.plot(roll_mad.roll_series, color=colors['m'])

    if roll_mad.test != -2:
        plt.axhline(y=mad_dict[roll_mad.test][0], color=colors['af'], linewidth=3)
        plt.axhline(y=mad_dict[roll_mad.test][1], color=colors['h2'], linewidth=3)
        plt.axhline(y=mad_dict[roll_mad.test][2], color=colors['s'], linewidth=3)

    if save_plot:
        if not save_plot_kwargs:
            save_plot_kwargs = {}
        plt.savefig(save_plot, **save_plot_kwargs)

    plt.show(block=False)



"""
Benford_py for Python is a module for application of Benford's Law
to a sequence of numbers.

Dependent on pandas, numpy and matplotlib

All logarithms ar in base 10: "log10"

Author:  Marcel Milcent

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .benford import *

__version__ = '0.4.1'



from pandas import DataFrame
from numpy import array, arange, log10
from .checks import _check_digs_
from .viz import plot_expected


class First(DataFrame):
    """Holds the expected probabilities of the First, First Two, or
    First Three digits according to Benford's distribution.

    Args:
        digs: 1, 2 or 3 - tells which of the first digits to consider:
            1 for the First Digit, 2 for the First Two Digits and 3 for
            the First Three Digits.
        plot: option to plot a bar chart of the Expected proportions.
            Defaults to True.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    """

    def __init__(self, digs, plot=True, save_plot=None, save_plot_kwargs=None):
        _check_digs_(digs)
        dig_name = f'First_{digs}_Dig'
        exp_array, dig_array = _gen_first_digits_(digs)
 
        DataFrame.__init__(self, {'Expected': exp_array}, index=dig_array)
        self.index.names = [dig_name]

        if plot:
            plot_expected(self, digs, save_plot=save_plot,
                          save_plot_kwargs=save_plot_kwargs)


class Second(DataFrame):
    """Holds the expected probabilities of the Second Digits
    according to Benford's distribution.

    Args:
        plot: option to plot a bar chart of the Expected proportions.
            Defaults to True.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    """
    def __init__(self, plot=True, save_plot=None, save_plot_kwargs=None):

        exp, sec_digs = _gen_second_digits_()

        DataFrame.__init__(self, {'Expected': exp, 'Sec_Dig': sec_digs})
        self.set_index("Sec_Dig", inplace=True)

        if plot:
            plot_expected(self, 22, save_plot=save_plot,
                          save_plot_kwargs=save_plot_kwargs)


class LastTwo(DataFrame):
    """Holds the expected probabilities of the Last Two Digits
    according to Benford's distribution.

    Args:
        plot: option to plot a bar chart of the Expected proportions.
            Defaults to True.
        save_plot: string with the path/name of the file in which the generated
            plot will be saved. Uses matplotlib.pyplot.savefig(). File format
            is infered by the file name extension. Only available when
            plot=True.
        save_plot_kwargs: dict with any of the kwargs accepted by
            matplotlib.pyplot.savefig()
            https://matplotlib.org/api/_as_gen/matplotlib.pyplot.savefig.html
            Only available when plot=True and save_plot is a string with the
            figure file path/name.
    """
    def __init__(self, num=False, plot=True, save_plot=None, save_plot_kwargs=None):
        exp, l2d = _gen_last_two_digits_(num=num)
        DataFrame.__init__(self, {'Expected': exp,
                                  'Last_2_Dig': l2d})
        self.set_index('Last_2_Dig', inplace=True)
        if plot:
            plot_expected(self, -2, save_plot=save_plot,
                          save_plot_kwargs=save_plot_kwargs)


def _get_expected_digits_(digs):
    """Chooses the Exxpected class to be used in a test

    Args:
        digs: the int corresponding to the Expected class to be instantiated

    Returns:
        the Expected instance forthe propoer test to be performed
    """
    if digs in [1, 2, 3]:
        return First(digs, plot=False)
    elif digs == 22:
        return Second(plot=False)
    else:
        return LastTwo(num=True, plot=False)


def _gen_last_two_digits_(num=False):
    """Creates two arrays, one with the possible last two digits and one with
    thei respective probabilities

    Args:
        num: returns numeric (ints) values. Defaluts to False,
            which returns strings.

    Returns:
        exp (np.array): Array with the (constant) probabilities of occurrence of
            each pair of last two digits 
        l2d (np.array): Array of ints or str, in any case representing all 100
            possible combinations of last two digits
    """
    exp = array([1 / 99.] * 100)
    l2d = arange(0, 100)
    if num:
        return exp, l2d
    l2d = l2d.astype(str)
    l2d[:10] = array(['00', '01', '02', '03', '04', '05',
                    '06', '07', '08', '09'])
    return exp, l2d

def _gen_first_digits_(digs):
    """Creates two arrays, one with the possible digits combinations and the
    other with their respective expected probabilities according to Benford

    Args:
        digs (int): 1, 2 or 3, for generation of the first, first two, or first
            three digits

    Returns:
        (tuple of arrays): the expected probabilities array and the digits
            combination array. 
    """
    dig_array = arange(10 ** (digs - 1), 10 ** digs)
    exp_prob = log10(1 + (1. / dig_array))
    return exp_prob, dig_array

def _gen_second_digits_():
    """Creates two arrays, one with he possible second digits combinations and
    the other with their respective expected probabilities according to Benford

    Returns:
        (tuple of arrays): the expected probabilities array and the second
        digits array.
    """
    exp_f2d, _ = _gen_first_digits_(2)
    sec_digs = range(10)
    sec_digs_in_f2d = array(list(range(10)) * 9)
    exp = array([exp_f2d[sec_digs_in_f2d == i].sum() for i in sec_digs])
    return exp, array(sec_digs)


import warnings
from .constants import mad_dict


def _inform_(df, high_Z, conf):
    """Selects and sorts by the Z_stats chosen to be considered, informing or not.
    """

    if isinstance(high_Z, int):
        if conf is not None:
            dd = df[['Expected', 'Found', 'Z_score'
                     ]].sort_values('Z_score', ascending=False).head(high_Z)
            print(f'\nThe entries with the top {high_Z} Z scores are:\n')
        # Summation Test
        else:
            dd = df[['Expected', 'Found', 'AbsDif'
                     ]].sort_values('AbsDif', ascending=False
                                    ).head(high_Z)
            print(f'\nThe entries with the top {high_Z} absolute deviations '
                  'are:\n')
    else:
        if high_Z == 'pos':
            m1 = df.Dif > 0
            m2 = df.Z_score > conf
            dd = df[['Expected', 'Found', 'Z_score'
                     ]].loc[m1 & m2].sort_values('Z_score', ascending=False)
            print('\nThe entries with the significant positive '
                  'deviations are:\n')
        elif high_Z == 'neg':
            m1 = df.Dif < 0
            m2 = df.Z_score > conf
            dd = df[['Expected', 'Found', 'Z_score'
                     ]].loc[m1 & m2].sort_values('Z_score', ascending=False)
            print('\nThe entries with the significant negative '
                  'deviations are:\n')
        else:
            dd = df[['Expected', 'Found', 'Z_score'
                     ]].loc[df.Z_score > conf].sort_values('Z_score',
                                                           ascending=False)
            print('\nThe entries with the significant deviations are:\n')
    print(dd)


def _report_mad_(digs, MAD):
    """Reports the test Mean Absolut Deviation and compares it to critical values
    """
    print(f'Mean Absolute Deviation: {MAD:.6f}')
    if digs != -2:
        mads = mad_dict[digs]
        if MAD <= mads[0]:
            print(f'MAD <= {mads[0]:.6f}: Close conformity.\n')
        elif MAD <= mads[1]:
            print(f'{mads[0]:.6f} < MAD <= {mads[1]:.6f}: '
                  'Acceptable conformity.\n')
        elif MAD <= mads[2]:
            print(f'{mads[1]:.6f} < MAD <= {mads[2]:.6f}: '
                  'Marginally Acceptable conformity.\n')
        else:
            print(f'MAD > {mads[2]:.6f}: Nonconformity.\n')
    else:
        print("There is no conformity check for this test's MAD.\n")


def _report_KS_(KS, crit_KS):
    """Reports the test Kolmogorov-Smirnov statistic and compares it to critical
    values, depending on the confidence level
    """
    result = 'PASS' if KS <= crit_KS else 'FAIL'
    print(f"\n\tKolmogorov-Smirnov: {KS:.6f}",
          f"\n\tCritical value: {crit_KS:.6f} -- {result}")


def _report_chi2_(chi2, crit_chi2):
    """Reports the test Chi-square statistic and compares it to critical values,
    depending on the confidence level
    """
    result = 'PASS' if chi2 <= crit_chi2 else 'FAIL'
    print(f"\n\tChi square: {chi2:.6f}",
          f"\n\tCritical value: {crit_chi2:.6f} -- {result}")


def _report_Z_(df, high_Z, crit_Z):
    """Reports the test Z scores and compares them to a critical value,
    depending on the confidence level
    """
    print(f"\n\tCritical Z-score:{crit_Z}.")
    _inform_(df, high_Z, crit_Z)


def _report_summ_(test, high_diff):
    """Reports the Summation Test Absolute Differences between the Found and
    the Expected proportions

    """
    if high_diff is not None:
        print(f'\nThe top {high_diff} Absolute Differences are:\n')
        print(test.sort_values('AbsDif', ascending=False).head(high_diff))
    else:
        print('\nThe top Absolute Differences are:\n')
        print(test.sort_values('AbsDif', ascending=False))


def _report_bhattac_coeff_(bhattac_coeff):
    """
    """
    print(f"Bhattacharyya Coefficient: {bhattac_coeff:6f}\n")


def _report_bhattac_dist_(bhattac_dist):
    """
    """
    print(f"Bhattacharyya Distance: {bhattac_dist:6f}\n")


def _report_kl_diverg_(kl_diverg):
    """
    """
    print(f"Kullback-Leibler Divergence: {kl_diverg:6f}\n")


def _report_test_(test, high=None, crit_vals=None):
    """Main report function. Receives the Args: to report with, initiates
    the process, and calls the right reporting helper function(s), depending
    on the Test.
    """
    print('\n', f'  {test.name}  '.center(50, '#'), '\n')
    if not 'Summation' in test.name:
        _report_mad_(test.digs, test.MAD)
        _report_bhattac_coeff_(test.bhattacharyya_coefficient)
        _report_bhattac_dist_(test.bhattacharyya_distance)
        _report_kl_diverg_(test.kullback_leibler_divergence)
        if test.confidence is not None:
            print(f"For confidence level {test.confidence}%: ")
            _report_KS_(test.KS, crit_vals['KS'])
            _report_chi2_(test.chi_square, crit_vals['chi2'])
            _report_Z_(test, high, crit_vals['Z'])
        else:
            print('Confidence is currently `None`. Set the confidence level, '
                  'so as to generate comparable critical values.')
            if isinstance(high, int):
                _inform_(test, high, None)
    else:
        _report_summ_(test, high)


def _report_mantissa_(stats):
    """Prints the mantissas statistics and their respective reference values

    Args:
        stats (dict): 
    """
    print("\n", '  Mantissas Test  '.center(52, '#'))
    print(f"\nThe Mantissas MEAN is      {stats['Mean']:.6f}."
          "\tRef: 0.5")
    print(f"The Mantissas VARIANCE is  {stats['Var']:.6f}."
          "\tRef: 0.08333")
    print(f"The Mantissas SKEWNESS is  {stats['Skew']:.6f}."
          "\tRef: 0.0")
    print(f"The Mantissas KURTOSIS is  {stats['Kurt']:.6f}."
          "\tRef: -1.2\n")


def _deprecate_inform_(verbose, inform):
    """
    Raises:
        FutureWarning: if the arg `inform` is used (to be deprecated).    
    """
    if inform is None:
        return verbose
    else:
        warnings.warn('The parameter `inform` will be deprecated in future '
                      'versions. Use `verbose` instead.',
                      FutureWarning)
        return inform



''' Setup for benford's module'''
from os import path
from setuptools import setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README-pypi.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='benford_py',
      version='0.4.1',
      description='A library for testing data sets with Bendford\'s Law',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/milcent/benford_py',
      download_url='https://github.com/milcent/benford_py/archive/v0.4.1.tar.gz',
      author='Marcel Milcent',
      author_email='marcelmilcent@gmail.com',
      license='BSD 3-Clause',
      packages=['benford'],
      install_requires=[
          'pandas',
          'numpy',
          'matplotlib',
      ],
      zip_safe=False,
      classifiers=[
          'Programming Language :: Python :: 3',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Development Status :: 2 - Pre-Alpha',
          'Intended Audience :: Financial and Insurance Industry',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Education',
          'Intended Audience :: Other Audience',
          'Topic :: Office/Business :: Financial :: Accounting',
          'Topic :: Scientific/Engineering :: Mathematics',
      ],)
